<!--
This is the directive for attaching machines to tokens.
-->
<div class="form-group">
    <label for="serial" translate>Machine</label>
    <input name="serial" type="text" ng-model="newMachine"
           autocomplete="new-password"
           placeholder="{{ 'start typing a machine.'|translate }}"
           typeahead-wait-ms="100"
           uib-typeahead="serial for serial in loadMachines($viewValue)"
           typeahead-no-results="noResults"
           typeahead-loading="loadingUsers" class="form-control" >
    <div ng-show="loadingMachines">
        <i class="glyphicon glyphicon-refresh"></i><span translate>Loading Machines...</span>
    </div>
    <div ng-show="noResults">
        <i class="glyphicon glyphicon-remove"></i><span translate>No machine found</span>
    </div>

</div>
