<!--
(c) NetKnights GmbH 2024,  https://netknights.it
This code is free software; you can redistribute it and/or
modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
as published by the Free Software Foundation; either
version 3 of the License, or any later version.
This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU AFFERO GENERAL PUBLIC LICENSE for more details.
You should have received a copy of the GNU Affero General Public
License along with this program.  If not, see <http://www.gnu.org/licenses/>.

SPDX-License-Identifier: AGPL-3.0-or-later
-->

<!-- Verify Token enrollment -->
<div ng-show="enrolledToken.rollout_state === 'verify'">
    <p translate>The token was enrolled, but you still need to verify it before you can use it!</p>
    <div class="form-group">
        <label for="verifyResponse">{{ enrolledToken.verify.message }}</label>
        <input type="text" ng-model="verifyResponse"
               id="verifyResponse"
               autofocus
               class="form-control"
               name="verifyResponse">
    </div>
    <div class="text-center">
        <button type="button" ng-click="sendVerifyResponse()"
                class="btn btn-primary" translate>Verify Token
        </button>
    </div>
</div>