


<ng-include src="'{{ instance }}/{{ "static/components/dialogs/views/dialog.about.html" | versioned }}'">
</ng-include>

<ng-include src="'{{ instance }}/{{ "static/components/dialogs/views/dialog.lock.html" | versioned }}'">
</ng-include>

<ng-include src="'{{ instance }}/{{ "static/components/dialogs/views/dialog.autocreate_realm.html" | versioned }}'">
</ng-include>

<ng-include src="'{{ instance }}/{{ "static/components/dialogs/views/dialog.welcome.html" | versioned }}'">
</ng-include>

<ng-include src="'{{ instance }}/{{ "static/components/dialogs/views/dialog.no.token.html" | versioned }}'">
</ng-include>

<!--load all javascripts after the HTML part-->
<script src="{{ instance }}/{{ "static/contrib/js/angular.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/contrib/js/ui-bootstrap-tpls.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/contrib/js/jquery.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/contrib/js/bootstrap.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/contrib/js/ngmodules/isteven-multi-select.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/contrib/js/ngmodules/ng-file-upload.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/contrib/js/hotkeys.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/contrib/js/ngmodules/angular-ui-router.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/contrib/js/ngmodules/angular-idle.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/contrib/js/ngmodules/angular-gettext.js" | versioned }}"></script>
<script src="{{ instance }}/{{  "static/components/translation/translations.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/contrib/js/ngmodules/angular-sanitize.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/app.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/providers/versioningProvider.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/providers/errorMessageProvider.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/directives/controllers/directives.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/login/factories/auth.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/login/factories/u2f.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/login/factories/webauthn.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/login/states/states.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/login/controllers/loginControllers.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/dashboard/controllers/dashboardControllers.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/dashboard/states/states.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/token/factories/token.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/token/factories/container.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/token/factories/validate.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/token/states/states.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/token/controllers/tokenControllers.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/token/controllers/tokenDetailController.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/token/controllers/tokenLostController.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/token/controllers/tokenGetSerialController.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/token/controllers/tokenChallengesController.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/token/controllers/tokenApplicationsController.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/token/controllers/containerControllers.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/token/controllers/containerTemplateControllers.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/user/factories/user.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/user/states/states.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/user/controllers/userControllers.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/audit/factories/audit.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/audit/states/states.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/audit/controllers/auditControllers.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/components/factories/component.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/components/states/states.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/components/controllers/componentControllers.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/config/factories/config.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/config/states/states.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/config/controllers/configControllers.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/config/controllers/realmController.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/config/controllers/eventController.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/config/controllers/periodicTaskController.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/config/controllers/ldapMachineResolverController.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/config/controllers/smtpServerController.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/config/controllers/radiusServerController.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/config/controllers/tokengroupController.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/config/controllers/serviceidController.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/config/controllers/privacyideaServerController.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/config/controllers/smsgatewayController.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/machine/factories/machine.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/machine/states/states.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/machine/controllers/machineController.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/machine/controllers/machineDetailsController.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/register/states/states.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/register/controllers/registerControllers.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/register/factories/registerFactory.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/recovery/states/states.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/recovery/controllers/recoveryControllers.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/recovery/factories/recoveryFactory.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/info/states/states.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/info/controllers/infoController.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/info/factories/info.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/components/filters/filters.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/contrib/js/angular-inform.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/contrib/js/webauthn-client/pi-webauthn.js" | versioned }}"></script>
<script src="{{ instance }}/{{ "static/contrib/js/u2f-api.js" | versioned }}"></script>
</body>
</html>
