<!--
(c) NetKnights GmbH 2024,  https://netknights.it
This code is free software; you can redistribute it and/or
modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
as published by the Free Software Foundation; either
version 3 of the License, or any later version.
This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU AFFERO GENERAL PUBLIC LICENSE for more details.
You should have received a copy of the GNU Affero General Public
License along with this program.  If not, see <http://www.gnu.org/licenses/>.

SPDX-FileCopyrightText: 2024 Jelina Unger <jelina.unger@netknights.it>
SPDX-License-Identifier: AGPL-3.0-or-later
-->

<table class="table table-hover">
    <tr>
        <td>
            <label translate>Passphrase (optional)</label>

            <div>
                <label for="passphrase_prompt" translate>
                    Passphrase Prompt
                </label>
                <div class="row">
                    <div class="col-sm-7">
                        <input class="form-control"
                               id="passphrase_prompt"
                               ng-model="passphrase.prompt"
                               placeholder="{{ 'This prompt will be displayed to the user.'|translate }}"
                               ng-disabled="passphrase.ad"
                               autofocus/>
                    </div>
                    <div class="col-sm-5">
                        <p class="help help-block" translate>
                            Enter a passphrase prompt that will be displayed to the user.
                            This should indicate which passphrase the user should enter.
                        </p>
                    </div>
                </div>
                <label for="passphrase_response" translate>
                    Passphrase Response
                </label>
                <div class="row">
                    <div class="col-sm-7">
                        <input class="form-control"
                               id="passphrase_response"
                               ng-model="passphrase.response"
                               placeholder="{{ 'Passphrase the user has to enter to successfully register the container'|translate }}"
                               ng-disabled="passphrase.ad"
                               autofocus/>
                    </div>
                    <div class="col-sm-5">
                        <p class="help help-block" translate>
                            This is the correct answer to the passphrase prompt the user has to enter to
                            successfully register the container.
                        </p>
                    </div>
                </div>
            </div>

        </td>
    </tr>
</table>