<div>
    <!-- Modal -->
    <div class="modal fade" id="dialogConfirmDelete" tabindex="-1"
         role="dialog"
         aria-labelledby="myModalConfirm"
         aria-hidden="true">
        <div class="modal-dialog">
            <div class="modal-content">
                <div class="modal-header">
                    <button type="button" class="close" data-dismiss="modal" aria-label="Close">
                        <span aria-hidden="true">&times;</span></button>
                    <h4 class="modal-title" id="myModalConfirm"
                        translate>Delete Configuration</h4>
                </div>
                <div class="modal-body">
                    <p>
                    {{ confirmDeleteObj.question }}
                    ({{ confirmDeleteObj.identifier }})</p>
                </div>

                <div class="modal-footer">
                    <button type="button" class="btn btn-default"
                            data-dismiss="modal" translate>No</button>
                    <button type="button"
                            class="btn btn-danger"
                            data-dismiss="modal"
                            ng-click="confirmDeleteObj.delete_function(confirmDeleteObj.identifier)">
                        <span class="glyphicon glyphicon-trash"></span>
                        <span translate>Delete</span>
                    </button>
                </div>
            </div>
        </div>
    </div>
</div>
