# (c) NetKnights GmbH 2024,  https://netknights.it
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-FileCopyrightText: 2024 Jelina Unger <jelina.unger@netknights.it>
# SPDX-License-Identifier: AGPL-3.0-or-later

class SmartphoneOptions:
    """
    Options for the smartphone container.
    """
    KEY_ALGORITHM = "key_algorithm"
    HASH_ALGORITHM = "hash_algorithm"
    ENCRYPT_KEY_ALGORITHM = "encrypt_key_algorithm"
    ENCRYPT_ALGORITHM = "encrypt_algorithm"
    ENCRYPT_MODE = "encrypt_mode"
