<div class="form-group col-sm-12">
<p class="help-block" translate>This documentation contains the current
    configuration of your system in restructured text. You can use
    this to document your installed system.</p>
<textarea ng-model="systemDocumentation"
          rows="15"
         class="form-group col-sm-12">
</textarea>
</div>
