<p class="help-block" translate>
    With the PUSH token privacyIDEA sends a notification to your registered smartphone asking
    if you want to approve the authentication step.
    You can accept or deny the authentication on your smartphone.
</p>
<p class="help-block" translate>
    For the enrollment process and for the authentication process your smartphone needs an internet connection
    to privacyIDEA.
</p>

<ng-include src="instanceUrl + '/static/components/token/views/token.display.apps.html'">
</ng-include>
