<p class="help-block" translate>
    The WebAuthn token is a token defined by the W3C and the Fido
    Alliance. You can register this token with any webservice and with
    as many web services you wish to.
</p>
