<div class="row">
    <div class="col-sm-6">
        <img width="{{ qrCodeWidth }}"
             class="centered"
             ng-src="{{ enrolledToken.otpkey.img }}"/>
    </div>
    <div class="col-sm-6">
        <span translate>Your Password:</span> <b>{{ enrolledToken.password }}</b>
    </div>

</div>
