class FIDO2TokenInfo:
    PUB_KEY = "pubKey"
    ORIGIN = "origin"
    AAGUID = "aaguid"
    ATTESTATION_LEVEL = "attestation_level"
    ATTESTATION_ISSUER = "attestation_issuer"
    ATTESTATION_SERIAL = "attestation_serial"
    ATTESTATION_SUBJECT = "attestation_subject"
    RELYING_PARTY_ID = "relying_party_id"
    RELYING_PARTY_NAME = "relying_party_name"
    CREDENTIAL_ID_HASH = "credential_id_hash"
    SIGN_COUNT = "sign_count"
    DEVICE_TYPE = "device_type"
    BACKED_UP = "backed_up"
    PUBLIC_KEY = "public_key"
    ATTESTATION_CERTIFICATE = "attestation_certificate"
    USER_ID = "fido2_user_id"
    CREDENTIAL_ID_HASH = "credential_id_hash"
    SIGN_COUNT = "sign_count"
    DEVICE_TYPE = "device_type"
    BACKED_UP = "backed_up"
    PUBLIC_KEY = "public_key"
    ATTESTATION_CERTIFICATE = "attestation_certificate"
    USER_ID = "fido2_user_id"