<div class="row">
    <div class="col-lg-3 sticky">
        <div class="pa-sidebar panel">
            <ul class="nav nav-pills nav-stacked nav-capitalize">
                <li role="presentation"
                    ng-show="checkRight('eventhandling_read')"
                    ng-class="{ active: $state.includes('config.events.list')}">
                    <a ui-sref="config.events.list" translate>
                        All Event Handlers</a></li>
                <li role="presentation"
                    ng-show="checkRight('eventhandling_write')"
                    ng-class="{ active: $state.includes('config.events.add')}">
                    <a ui-sref="config.events.add"
                       translate>Create new Event Handler</a></li>
                <hr>
                <li role="presentation">
                    <a target="_external" ng-href="{{ 'https://privacyidea.readthedocs.io/en/v'+
                    privacyideaVersionNumber+'/eventhandler/index.html'}}">
                        <span class="glyphicon glyphicon-info-sign"></span>
                        <span translate>Help about Events</span></a></li>
            </ul>
        </div>

    </div>
    <div ui-view class="col-lg-9 slide panel">
    </div>
</div>
