<h2 translate>DayPassword Token settings</h2>
<p class="help-block" translate>
    The DayPassword Token is a time based password token.
    It is based on the TOTP token, but it can be used more than once.</a>.
</p>
<p translate>
    Here you can define settings, that will be set as default values, when
    enrolling a DayPassword token. Note that the user/admin policies <code>daypassword_hashlib</code> and
    <code>daypassword_timestep</code> always have precedence over these default values.
</p>

<div class="form-group">
    <label for="daypasswordStep" translate>Default Time Step</label>
    <input type="text"
           required
           class="form-control"
           placeholder="60"
           ng-model="form['daypassword.timeStep']" name="daypasswordStep">
</div>

<div class="form-group">
    <label for="daypasswordHashlib" translate>Default Hashlib</label>
    <select class="form-control"
           ng-model="form['daypassword.hashlib']" name="daypasswordHashlib"
        ng-options="select for select in formInit.hashlibs">
    </select>
</div>
