{% include 'header.html' %}

<!-- Include menu -->

<ng-include src="'{{ instance }}/static/{{ customization_menu_file }}'">
</ng-include>
<div class="container">
    <div ui-view class="slide"></div>
</div>

<input type=hidden id=REMOTE_USER value="{{ remote_user }}">
<input type=hidden id=FORCE_REMOTE_USER value="{{ force_remote_user }}">
<input type=hidden id=PASSWORD_RESET value="{{ password_reset }}">
<input type=hidden id=HSM_READY value="{{ hsm_ready }}">
<input type=hidden id=CUSTOMIZATION value="{{ customization }}">
<input type=hidden id="REALMS" value="{{ realms }}">
<input type=hidden id="LOGO" value="{{ logo }}">
<input type=hidden id="SHOW_NODE" value="{{ show_node }}">
<input type=hidden id="EXTERNAL_LINKS" value="{{ external_links }}">
<input type=hidden id="HAS_JOB_QUEUE" value="{{ has_job_queue }}">
<input type=hidden id="LOGIN_TEXT" value="{{ login_text }}">
<input type=hidden id="LOGOUT_REDIRECT_URL" value="{{ logout_redirect_url }}">
<input type=hidden id="GDPR_LINK" value="{{ gdpr_link }}">
<input type=hidden id="PRIVACYIDEA_VERSION_NUMBER" value="{{ privacyideaVersionNumber }}">
<input type=hidden id=PI_TRANSLATION_WARNING value="{{ translation_warning }}">

<!-- Include baseline -->

<ng-include src="'{{ instance }}/static/{{ customization_baseline_file }}'">
</ng-include>

{% include 'footer.html' %}
