<p class="help-block" translate>
    The indexed secret Token is based on a shared secret between privacyIDEA
    and the user.
    During authentication the user is asked for random positions from this known
    secret.
</p>

<h4 translate>Token data</h4>

<div class="form-group" ng-hide="tokensettings.indexedsecret.force_attibute">
    <label for="otpkey" translate>Secret</label>
    <input type="text" class="form-control"
            id="otpkey"
            placeholder="{{ 'Your secret phrase.'|translate }}"
            ng-model="form.otpkey" />
</div>

<div class="help-block" translate ng-show="tokensettings.indexedsecret.force_attibute">
    The value of the indexed secret will be set by a policy.
</div>
