<div id="baseline" class="navbar navbar-inverse navbar-fixed-bottom">
    <div class="container">
        <a ng-click="about()"
            class="navbar-text pull-left">
            privacyIDEA
            {{ privacyideaVersionNumber }}
        </a>


        <a class="navbar-btn btn btn-success pull-right"
           ng-show="piExternalLinks !== 'False'"
           ng-href="{{ privacyideaSupportLink }}"
           target="_external">
            <translate>Support</translate>
        </a>
        <a class="navbar-text pull-right"
           ng-show="piExternalLinks !== 'False'"
                href="https://community.privacyidea.org"
                target="_external">
            privacyIDEA Forum</a>

        <a class="navbar-text pull-right"
           ng-show="privacyideaVersionNumber"
                ng-href="https://privacyidea.readthedocs.io/en/v{{ privacyideaVersionNumber }}"
                target="documentation">
            <translate>Online Documentation</translate>
        </a>
       <a class="navbar-text pull-right pull-right"
           ng-show="piGDPRLink"
           ng-href="{{ piGDPRLink }}"
           target="_external">
            <translate>Privacy Statement</translate>
        </a>
    </div>
</div>
